require 'json'
require 'logger'
require 'sinatra/base'

class Driftwood < Sinatra::Base
  require 'driftwood/version'
  require 'driftwood/slack'
  require 'driftwood/bigquery'
  require 'driftwood/monkeypatches'
  require 'driftwood/plugin'

  set :logging, true
  set :strict,  true
  set :views,         File.dirname(__FILE__) + '/../views'
  set :public_folder, File.dirname(__FILE__) + '/../public'

  # this allows other tools to register handlers and interact with the db
  attr_reader :slack, :bigquery

  def initialize(app=nil)
    super(app)
    $logger.info "Starting Driftwood v#{Driftwood::VERSION}"
    @bigquery = Driftwood::Bigquery.new(settings.gcloud)
    @slack    = Driftwood::Slack.new(settings.slack)
    @app_id   = settings.slack[:client_id]
    @slack.authorize(cached_auth)

    @plugins = settings.plugins.map do |plugin, config|
      filename = plugin.snake_case

      begin
        require "driftwood/plugin/#{filename}"
      rescue LoadError
        require filename
      end

      Driftwood::Plugin::const_get(plugin).new(config, @slack, @bigquery)
    end

    # help handler
    @slack.register_handler('message') do |team_id, event_data|
      user = event_data['user']
      text = event_data['text']

      case text
      when /^help$/i
        @slack.send_response(team_id, user, "Usage for all plugins:")
        @plugins.each do |plugin|
          next unless plugin.usage
          @slack.send_response(team_id, user, plugin.usage)
        end
      when /^help (\w*)$/i
        name = $1
        @slack.send_response(team_id, user, "Usage for #{name}:")
        @plugins.each do |plugin|
          next unless plugin.name = name.downcase
          @slack.send_response(team_id, user, plugin.usage)
        end
      end
    end

  end

  get '/' do
    if params[:code]
      begin
        save_auth_cache(@slack.authorize(params[:code]))
        @message = "Rad, you've been authenticated!"
        status 200

      rescue => e
        $logger.error e.message
        $logger.debug e.backtrace.join("\n")

        status 403
        @error = e.message
      end
    end

    erb :index
  end

  post '/events' do
    begin
      # Return HTTP status code 200 so Slack knows we've received the Event
      status 200
      body @slack.event(JSON.parse(request.body.read))
    rescue => e
      $logger.error e.message
      $logger.debug e.backtrace.join("\n")
      status 403
      body "Error: #{e.message}"
    end
  end


  not_found do
    halt 404, "You shall not pass! (page not found)\n"
  end

  helpers do
    def cached_auth
      @bigquery.get_auth_tokens(@app_id)
    end

    def save_auth_cache(auth)
      @bigquery.insert_team(auth, @app_id)
    end
  end

end
