class Driftwood::Plugin::Logger < Driftwood::Plugin

  def initialize(config=nil, slack=nil, bigquery=nil)
    super(config, slack, bigquery)
    @about   = 'Logs all public messages to BigQuery.'

    # logging handler
    @slack.register_handler('message') do |team_id, event_data|
      next unless event_data['channel_type'] == 'channel'
      $logger.debug "Logged message for #{team_id}: #{event_data['text']}"
      @bigquery.insert_message(@slack.normalize_message(team_id, event_data))
    end

    # register new users
    @slack.register_handler('team_join') do |team_id, event_data|
      user = @slack.normalize_user(event_data['user'])
      $logger.info "Registered new user: #{user['name']} (#{user['real_name']})"
      @bigquery.insert_user(user)
    end

    # Do this in a thread so it doesn't block initialization
    # These methods may take a long time to complete
    Thread.new do
      begin
        $logger.info "Starting data synchronization..."
        sync_messages
        sync_channels
        sync_users
      rescue => e
        $logger.error e.message
        $logger.debug e.backtrace.join("\n")
      end

      loop do
        sleep 86400 # Once a day; 60 * 60 * 24
        begin
          $logger.info "Purging deleted users..."
          @bigquery.purge_deleted_users
        rescue => e
          $logger.error e.message
          $logger.debug e.backtrace.join("\n")
        end
      end
    end
  end

  def sync_channels
    @slack.all_channels.each do |channel|
      @bigquery.insert_channel(channel)
    end
    $logger.info "All channels synced."
  end

  def sync_users
    @slack.all_users.each do |user|
      @bigquery.insert_user(user)
    end
    @bigquery.reconcile_user_creations
    $logger.info "All users synced."
  end

  def sync_messages
    starting_from = @bigquery.newest_message_timestamp
    @slack.all_messages(starting_from).each do |message|
      @bigquery.insert_message(message)
    end
    $logger.info "All messages synced."
  end

end